"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)
    tests.unit.utils.cloud_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Test the salt-cloud utilities module.

"""


import os
import tempfile

import pytest

import salt.utils.cloud as cloud
from salt.utils.cloud import __ssh_gateway_arguments as ssh_gateway_arguments
from tests.support.mock import MagicMock, patch


@pytest.fixture()
def configure_loader_modules():
    return {
        cloud: {"__opts__": {"sock_dir": "master"}},
    }


@pytest.fixture
def create_class(tmp_path):
    old_cwd = os.getcwd()
    gpg_keydir = tmp_path / "gpg-keydir"
    gpg_keydir.mkdir()
    try:
        # The keyring library uses `getcwd()`, let's make sure we in a good directory
        # before importing keyring
        os.chdir(str(gpg_keydir))
        # Late import because of the above reason
        import keyring
        import keyring.backend

        class CustomKeyring(keyring.backend.KeyringBackend):
            """
            A test keyring which always outputs same password
            """

            def __init__(self):
                self.__storage = {}

            def supported(self):
                return 0

            def set_password(
                self, servicename, username, password
            ):  # pylint: disable=arguments-differ
                self.__storage.setdefault(servicename, {}).update({username: password})
                return 0

            def get_password(
                self, servicename, username
            ):  # pylint: disable=arguments-differ
                return self.__storage.setdefault(servicename, {}).get(username)

            def delete_password(
                self, servicename, username
            ):  # pylint: disable=arguments-differ
                self.__storage.setdefault(servicename, {}).pop(username, None)
                return 0

        # set the keyring for keyring lib
        keyring.set_keyring(CustomKeyring())
        yield
    except ImportError:
        pytest.skip('The "keyring" python module is not installed')
    finally:
        os.chdir(old_cwd)


def test_ssh_password_regex(create_class):
    """Test matching ssh password patterns"""
    for pattern in (
        "Password for root@127.0.0.1:",
        "root@127.0.0.1 Password:",
        " Password:",
    ):
        assert cloud.SSH_PASSWORD_PROMP_RE.match(pattern) is not None
        assert cloud.SSH_PASSWORD_PROMP_RE.match(pattern.lower()) is not None
        assert cloud.SSH_PASSWORD_PROMP_RE.match(pattern.strip()) is not None
        assert cloud.SSH_PASSWORD_PROMP_RE.match(pattern.lower().strip()) is not None


def test__save_password_in_keyring(create_class):
    """
    Test storing password in the keyring
    """
    # Late import
    import keyring

    cloud._save_password_in_keyring(
        "salt.cloud.provider.test_case_provider",
        "fake_username",
        "fake_password_c8231",
    )
    stored_pw = keyring.get_password(
        "salt.cloud.provider.test_case_provider",
        "fake_username",
    )
    keyring.delete_password(
        "salt.cloud.provider.test_case_provider",
        "fake_username",
    )
    assert stored_pw == "fake_password_c8231"


def test_retrieve_password_from_keyring(create_class):
    # Late import
    import keyring

    keyring.set_password(
        "salt.cloud.provider.test_case_provider",
        "fake_username",
        "fake_password_c8231",
    )
    pw_in_keyring = cloud.retrieve_password_from_keyring(
        "salt.cloud.provider.test_case_provider", "fake_username"
    )
    assert pw_in_keyring == "fake_password_c8231"


def test_sftp_file_with_content_under_python3(create_class):
    with pytest.raises(Exception) as context:
        cloud.sftp_file("/tmp/test", "ТЕSТ test content")
        # we successful pass the place with os.write(tmpfd, ...
        assert "a bytes-like object is required, not 'str'" != str(context.exception)


@pytest.mark.skip_on_windows(reason="Not applicable for Windows.")
def test_check_key_path_and_mode(create_class):
    with tempfile.NamedTemporaryFile() as f:
        key_file = f.name

        os.chmod(key_file, 0o644)
        assert cloud.check_key_path_and_mode("foo", key_file) is False
        os.chmod(key_file, 0o600)
        assert cloud.check_key_path_and_mode("foo", key_file) is True
        os.chmod(key_file, 0o400)
        assert cloud.check_key_path_and_mode("foo", key_file) is True

    # tmp file removed
    assert cloud.check_key_path_and_mode("foo", key_file) is False


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_default_port():
    """
    Test deploy_windows with default port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    with patch("salt.utils.smb.get_conn", MagicMock()) as mock, patch(
        "salt.utils.smb.mkdirs", MagicMock()
    ), patch("salt.utils.smb.put_file", MagicMock()), patch(
        "salt.utils.smb.delete_file", MagicMock()
    ), patch(
        "salt.utils.smb.delete_directory", MagicMock()
    ), patch(
        "time.sleep", MagicMock()
    ), patch.object(
        cloud, "wait_for_port", mock_true
    ), patch.object(
        cloud, "fire_event", MagicMock()
    ), patch.object(
        cloud, "wait_for_psexecsvc", mock_true
    ), patch.object(
        cloud, "run_psexec_command", mock_tuple
    ):

        cloud.deploy_windows(host="test", win_installer="")
        mock.assert_called_once_with("test", "Administrator", None, 445)


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_custom_port():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    with patch("salt.utils.smb.get_conn", MagicMock()) as mock, patch(
        "salt.utils.smb.mkdirs", MagicMock()
    ), patch("salt.utils.smb.put_file", MagicMock()), patch(
        "salt.utils.smb.delete_file", MagicMock()
    ), patch(
        "salt.utils.smb.delete_directory", MagicMock()
    ), patch(
        "time.sleep", MagicMock()
    ), patch.object(
        cloud, "wait_for_port", mock_true
    ), patch.object(
        cloud, "fire_event", MagicMock()
    ), patch.object(
        cloud, "wait_for_psexecsvc", mock_true
    ), patch.object(
        cloud, "run_psexec_command", mock_tuple
    ):

        cloud.deploy_windows(host="test", port=1234, win_installer="")
        mock.assert_called_once_with("test", "Administrator", None, 1234)


def test_run_psexec_command_cleanup_lingering_paexec():
    pytest.importorskip("pypsexec.client", reason="Requires PyPsExec")
    mock_psexec = patch("salt.utils.cloud.PsExecClient", autospec=True)
    mock_scmr = patch("salt.utils.cloud.ScmrService", autospec=True)
    # We're mocking 'remove_service' because all we care about is the cleanup
    # command
    mock_rm_svc = patch("salt.utils.cloud.Client.remove_service", autospec=True)
    with mock_scmr, mock_rm_svc, mock_psexec as mock_client:
        mock_client.return_value.session = MagicMock(username="Gary")
        mock_client.return_value.connection = MagicMock(server_name="Krabbs")
        mock_client.return_value.run_executable.return_value = (
            "Sandy",
            "MermaidMan",
            "BarnicleBoy",
        )
        cloud.run_psexec_command(
            "spongebob",
            "squarepants",
            "patrick",
            "squidward",
            "plankton",
        )
        mock_client.return_value.cleanup.assert_called_once()


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_programdata():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    mock_conn = MagicMock()

    with patch("salt.utils.smb", MagicMock()) as mock_smb:
        mock_smb.get_conn.return_value = mock_conn
        mock_smb.mkdirs.return_value = None
        mock_smb.put_file.return_value = None
        mock_smb.delete_file.return_value = None
        mock_smb.delete_directory.return_value = None
        with patch("time.sleep", MagicMock()), patch.object(
            cloud, "wait_for_port", mock_true
        ), patch.object(cloud, "fire_event", MagicMock()), patch.object(
            cloud, "wait_for_psexecsvc", mock_true
        ), patch.object(
            cloud, "run_psexec_command", mock_tuple
        ):
            cloud.deploy_windows(host="test", win_installer="")
            expected = "ProgramData/Salt Project/Salt/conf/pki/minion"
            mock_smb.mkdirs.assert_called_with(expected, conn=mock_conn)


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_programdata_minion_pub():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    mock_conn = MagicMock()

    with patch("salt.utils.smb", MagicMock()) as mock_smb:
        mock_smb.get_conn.return_value = mock_conn
        mock_smb.mkdirs.return_value = None
        mock_smb.put_file.return_value = None
        mock_smb.put_str.return_value = None
        mock_smb.delete_file.return_value = None
        mock_smb.delete_directory.return_value = None
        with patch("time.sleep", MagicMock()), patch.object(
            cloud, "wait_for_port", mock_true
        ), patch.object(cloud, "fire_event", MagicMock()), patch.object(
            cloud, "wait_for_psexecsvc", mock_true
        ), patch.object(
            cloud, "run_psexec_command", mock_tuple
        ):
            cloud.deploy_windows(host="test", minion_pub="pub", win_installer="")
            expected = "ProgramData\\Salt Project\\Salt\\conf\\pki\\minion\\minion.pub"
            mock_smb.put_str.assert_called_with("pub", expected, conn=mock_conn)


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_programdata_minion_pem():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    mock_conn = MagicMock()

    with patch("salt.utils.smb", MagicMock()) as mock_smb:
        mock_smb.get_conn.return_value = mock_conn
        mock_smb.mkdirs.return_value = None
        mock_smb.put_file.return_value = None
        mock_smb.put_str.return_value = None
        mock_smb.delete_file.return_value = None
        mock_smb.delete_directory.return_value = None
        with patch("time.sleep", MagicMock()), patch.object(
            cloud, "wait_for_port", mock_true
        ), patch.object(cloud, "fire_event", MagicMock()), patch.object(
            cloud, "wait_for_psexecsvc", mock_true
        ), patch.object(
            cloud, "run_psexec_command", mock_tuple
        ):
            cloud.deploy_windows(host="test", minion_pem="pem", win_installer="")
            expected = "ProgramData\\Salt Project\\Salt\\conf\\pki\\minion\\minion.pem"
            mock_smb.put_str.assert_called_with("pem", expected, conn=mock_conn)


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_programdata_master_sign_pub_file():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    mock_conn = MagicMock()

    with patch("salt.utils.smb", MagicMock()) as mock_smb:
        mock_smb.get_conn.return_value = mock_conn
        mock_smb.mkdirs.return_value = None
        mock_smb.put_file.return_value = None
        mock_smb.put_str.return_value = None
        mock_smb.delete_file.return_value = None
        mock_smb.delete_directory.return_value = None
        with patch("time.sleep", MagicMock()), patch.object(
            cloud, "wait_for_port", mock_true
        ), patch.object(cloud, "fire_event", MagicMock()), patch.object(
            cloud, "wait_for_psexecsvc", mock_true
        ), patch.object(
            cloud, "run_psexec_command", mock_tuple
        ):
            cloud.deploy_windows(
                host="test", master_sign_pub_file="test.txt", win_installer=""
            )
            expected = (
                "ProgramData\\Salt Project\\Salt\\conf\\pki\\minion\\master_sign.pub"
            )
            called = False
            for call in mock_smb.put_file.mock_calls:
                if expected in call[1]:
                    called = True
            assert called


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_programdata_minion_conf_grains():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    mock_conn = MagicMock()

    with patch("salt.utils.smb", MagicMock()) as mock_smb:
        mock_smb.get_conn.return_value = mock_conn
        mock_smb.mkdirs.return_value = None
        mock_smb.put_file.return_value = None
        mock_smb.put_str.return_value = None
        mock_smb.delete_file.return_value = None
        mock_smb.delete_directory.return_value = None
        with patch("time.sleep", MagicMock()), patch.object(
            cloud, "wait_for_port", mock_true
        ), patch.object(cloud, "fire_event", MagicMock()), patch.object(
            cloud, "wait_for_psexecsvc", mock_true
        ), patch.object(
            cloud, "run_psexec_command", mock_tuple
        ):
            minion_conf = {"grains": {"spongebob": "squarepants"}}
            cloud.deploy_windows(host="test", minion_conf=minion_conf, win_installer="")
            expected = "ProgramData\\Salt Project\\Salt\\conf\\grains"
            called = False
            for call in mock_smb.put_str.mock_calls:
                if expected in call[1]:
                    called = True
            assert called


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_deploy_windows_programdata_minion_conf():
    """
    Test deploy_windows with a custom port
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    mock_conn = MagicMock()

    with patch("salt.utils.smb", MagicMock()) as mock_smb:
        mock_smb.get_conn.return_value = mock_conn
        mock_smb.mkdirs.return_value = None
        mock_smb.put_file.return_value = None
        mock_smb.put_str.return_value = None
        mock_smb.delete_file.return_value = None
        mock_smb.delete_directory.return_value = None
        with patch("time.sleep", MagicMock()), patch.object(
            cloud, "wait_for_port", mock_true
        ), patch.object(cloud, "fire_event", MagicMock()), patch.object(
            cloud, "wait_for_psexecsvc", mock_true
        ), patch.object(
            cloud, "run_psexec_command", mock_tuple
        ):
            minion_conf = {"master": "test-master"}
            cloud.deploy_windows(host="test", minion_conf=minion_conf, win_installer="")
            config = (
                "ipc_mode: tcp\r\n"
                "master: test-master\r\n"
                "multiprocessing: true\r\n"
                "pki_dir: /conf/pki/minion\r\n"
            )
            expected = "ProgramData\\Salt Project\\Salt\\conf\\minion"
            mock_smb.put_str.assert_called_with(config, expected, conn=mock_conn)


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
def test_winrm_pinnned_version():
    """
    Test that winrm is pinned to a version 0.3.0 or higher.
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    with patch("salt.utils.smb.get_conn", MagicMock()), patch(
        "salt.utils.smb.mkdirs", MagicMock()
    ), patch("salt.utils.smb.put_file", MagicMock()), patch(
        "salt.utils.smb.delete_file", MagicMock()
    ), patch(
        "salt.utils.smb.delete_directory", MagicMock()
    ), patch(
        "time.sleep", MagicMock()
    ), patch.object(
        cloud, "wait_for_port", mock_true
    ), patch.object(
        cloud, "fire_event", MagicMock()
    ), patch.object(
        cloud, "wait_for_psexecsvc", mock_true
    ), patch.object(
        cloud, "run_psexec_command", mock_tuple
    ):

        try:
            import winrm
        except ImportError:
            raise pytest.skip('The "winrm" python module is not installed in this env.')
        else:
            import pkg_resources

            winrm_pkg = pkg_resources.get_distribution("pywinrm")
            assert winrm_pkg.version >= "0.3.0"
    # fmt: on


def test_ssh_gateway_arguments_default_alive_args():
    server_alive_interval = 60
    server_alive_count_max = 3
    arguments = ssh_gateway_arguments({"ssh_gateway": "host"})
    assert "-oServerAliveInterval={}".format(server_alive_interval) in arguments
    assert "-oServerAliveCountMax={}".format(server_alive_count_max) in arguments


def test_ssh_gateway_arguments_alive_args():
    server_alive_interval = 10
    server_alive_count_max = 8
    arguments = ssh_gateway_arguments(
        {
            "ssh_gateway": "host",
            "server_alive_interval": server_alive_interval,
            "server_alive_count_max": server_alive_count_max,
        }
    )
    assert "-oServerAliveInterval={}".format(server_alive_interval) in arguments
    assert "-oServerAliveCountMax={}".format(server_alive_count_max) in arguments


def test_wait_for_port_default_alive_args():
    server_alive_interval = 60
    server_alive_count_max = 3
    with patch("salt.utils.cloud.socket", autospec=True), patch(
        "salt.utils.cloud._exec_ssh_cmd", autospec=True, return_value=0
    ) as exec_ssh_cmd:
        cloud.wait_for_port(
            "127.0.0.1",
            gateway={"ssh_gateway": "host", "ssh_gateway_user": "user"},
        )
        assert exec_ssh_cmd.call_count == 2
        ssh_call = exec_ssh_cmd.call_args[0][0]
        assert "-oServerAliveInterval={}".format(server_alive_interval) in ssh_call
        assert "-oServerAliveCountMax={}".format(server_alive_count_max) in ssh_call


def test_wait_for_port_alive_args():
    server_alive_interval = 66
    server_alive_count_max = 1
    with patch("salt.utils.cloud.socket", autospec=True), patch(
        "salt.utils.cloud._exec_ssh_cmd", autospec=True, return_value=0
    ) as exec_ssh_cmd:
        cloud.wait_for_port(
            "127.0.0.1",
            server_alive_interval=server_alive_interval,
            server_alive_count_max=server_alive_count_max,
            gateway={"ssh_gateway": "host", "ssh_gateway_user": "user"},
        )
        assert exec_ssh_cmd.call_count == 2
        ssh_call = exec_ssh_cmd.call_args[0][0]
        assert "-oServerAliveInterval={}".format(server_alive_interval) in ssh_call
        assert "-oServerAliveCountMax={}".format(server_alive_count_max) in ssh_call


def test_scp_file_default_alive_args():
    server_alive_interval = 60
    server_alive_count_max = 3
    with patch("salt.utils.cloud.socket", autospec=True), patch(
        "salt.utils.cloud._exec_ssh_cmd", autospec=True, return_value=0
    ) as exec_ssh_cmd:
        cloud.scp_file(
            "/salt.txt",
            contents=None,
            kwargs={"hostname": "127.0.0.1", "username": "user"},
            local_file="/salt.txt",
        )
        assert exec_ssh_cmd.call_count == 1
        ssh_call = exec_ssh_cmd.call_args[0][0]
        assert "-oServerAliveInterval={}".format(server_alive_interval) in ssh_call
        assert "-oServerAliveCountMax={}".format(server_alive_count_max) in ssh_call


def test_scp_file_alive_args():
    server_alive_interval = 64
    server_alive_count_max = 4
    with patch("salt.utils.cloud.socket", autospec=True), patch(
        "salt.utils.cloud._exec_ssh_cmd", autospec=True, return_value=0
    ) as exec_ssh_cmd:
        cloud.scp_file(
            "/salt.txt",
            contents=None,
            kwargs={
                "hostname": "127.0.0.1",
                "username": "user",
                "server_alive_interval": server_alive_interval,
                "server_alive_count_max": server_alive_count_max,
            },
            local_file="/salt.txt",
        )
        assert exec_ssh_cmd.call_count == 1
        ssh_call = exec_ssh_cmd.call_args[0][0]
        assert "-oServerAliveInterval={}".format(server_alive_interval) in ssh_call
        assert "-oServerAliveCountMax={}".format(server_alive_count_max) in ssh_call


def test_sftp_file_default_alive_args():
    server_alive_interval = 60
    server_alive_count_max = 3
    with patch("salt.utils.cloud.socket", autospec=True), patch(
        "salt.utils.cloud._exec_ssh_cmd", autospec=True, return_value=0
    ) as exec_ssh_cmd:
        cloud.sftp_file(
            "/salt.txt",
            contents=None,
            kwargs={"hostname": "127.0.0.1", "username": "user"},
            local_file="/salt.txt",
        )
        assert exec_ssh_cmd.call_count == 1
        ssh_call = exec_ssh_cmd.call_args[0][0]
        assert "-oServerAliveInterval={}".format(server_alive_interval) in ssh_call
        assert "-oServerAliveCountMax={}".format(server_alive_count_max) in ssh_call


def test_sftp_file_alive_args():
    server_alive_interval = 62
    server_alive_count_max = 6
    with patch("salt.utils.cloud.socket", autospec=True), patch(
        "salt.utils.cloud._exec_ssh_cmd", autospec=True, return_value=0
    ) as exec_ssh_cmd:
        cloud.sftp_file(
            "/salt.txt",
            contents=None,
            kwargs={
                "hostname": "127.0.0.1",
                "username": "user",
                "server_alive_interval": server_alive_interval,
                "server_alive_count_max": server_alive_count_max,
            },
            local_file="/salt.txt",
        )
        assert exec_ssh_cmd.call_count == 1
        ssh_call = exec_ssh_cmd.call_args[0][0]
        assert "-oServerAliveInterval={}".format(server_alive_interval) in ssh_call
        assert "-oServerAliveCountMax={}".format(server_alive_count_max) in ssh_call


def test_deploy_script_ssh_timeout():
    with patch("salt.utils.cloud.root_cmd", return_value=False) as root_cmd, patch(
        "salt.utils.cloud.wait_for_port", return_value=True
    ), patch("salt.utils.cloud.wait_for_passwd", return_value=True), patch(
        "salt.utils.cloud._exec_ssh_cmd"
    ):
        cloud.deploy_script("127.0.0.1", ssh_timeout=34)
        # verify that ssh_timeout made it into ssh_kwargs
        assert root_cmd.call_count == 1
        ssh_kwargs = root_cmd.call_args.kwargs
        assert "ssh_timeout" in ssh_kwargs
        assert ssh_kwargs["ssh_timeout"] == 34


@pytest.mark.parametrize(
    "master,expected",
    [
        (None, None),
        ("single_master", "single_master"),
        (["master1", "master2", "master3"], "master1,master2,master3"),
    ],
)
def test__format_master_param(master, expected):
    result = cloud._format_master_param(master)
    assert result == expected


@pytest.mark.skip_unless_on_windows(reason="Only applicable for Windows.")
@pytest.mark.parametrize(
    "master,expected",
    [
        (None, None),
        ("single_master", "single_master"),
        (["master1", "master2", "master3"], "master1,master2,master3"),
    ],
)
def test_deploy_windows_master(master, expected):
    """
    Test deploy_windows with master parameter
    """
    mock_true = MagicMock(return_value=True)
    mock_tuple = MagicMock(return_value=(0, 0, 0))
    with patch("salt.utils.smb.get_conn", MagicMock()), patch(
        "salt.utils.smb.mkdirs", MagicMock()
    ), patch("salt.utils.smb.put_file", MagicMock()), patch(
        "salt.utils.smb.delete_file", MagicMock()
    ), patch(
        "salt.utils.smb.delete_directory", MagicMock()
    ), patch(
        "time.sleep", MagicMock()
    ), patch.object(
        cloud, "wait_for_port", mock_true
    ), patch.object(
        cloud, "fire_event", MagicMock()
    ), patch.object(
        cloud, "wait_for_psexecsvc", mock_true
    ), patch.object(
        cloud, "run_psexec_command", mock_tuple
    ) as mock:
        cloud.deploy_windows(host="test", win_installer="install.exe", master=master)
        expected_cmd = "c:\\salttemp\\install.exe"
        expected_args = "/S /master={} /minion-name=None".format(expected)
        assert mock.call_args_list[0].args[0] == expected_cmd
        assert mock.call_args_list[0].args[1] == expected_args



"""
Beacon to emit Twilio text messages
"""
import logging

import salt.utils.beacons

try:
    import twilio

    # Grab version, ensure elements are ints
    twilio_version = tuple(int(x) for x in twilio.__version_info__)
    if twilio_version > (5,):
        from twilio.rest import Client as TwilioRestClient
    else:
        from twilio.rest import TwilioRestClient
    HAS_TWILIO = True
except ImportError:
    HAS_TWILIO = False

log = logging.getLogger(__name__)

__virtualname__ = "twilio_txt_msg"


def __virtual__():
    if HAS_TWILIO:
        return __virtualname__
    else:
        err_msg = "twilio library is missing."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for twilio_txt_msg beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for twilio_txt_msg beacon must be a list."
    else:
        config = salt.utils.beacons.list_to_dict(config)

        if not all(x in config for x in ("account_sid", "auth_token", "twilio_number")):
            return (
                False,
                "Configuration for twilio_txt_msg beacon "
                "must contain account_sid, auth_token "
                "and twilio_number items.",
            )
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Emit a dict name "texts" whose value is a list
    of texts.

    .. code-block:: yaml

        beacons:
          twilio_txt_msg:
            - account_sid: "<account sid>"
            - auth_token: "<auth token>"
            - twilio_number: "+15555555555"
            - interval: 10

    """
    log.trace("twilio_txt_msg beacon starting")

    config = salt.utils.beacons.list_to_dict(config)

    ret = []
    if not all([config["account_sid"], config["auth_token"], config["twilio_number"]]):
        return ret
    output = {}
    output["texts"] = []
    client = TwilioRestClient(config["account_sid"], config["auth_token"])
    messages = client.messages.list(to=config["twilio_number"])
    log.trace("Num messages: %d", len(messages))
    if not messages:
        log.trace("Twilio beacon has no texts")
        return ret

    for message in messages:
        item = {}
        item["id"] = str(message.sid)
        item["body"] = str(message.body)
        item["from"] = str(message.from_)
        item["sent"] = str(message.date_sent)
        item["images"] = []

        if int(message.num_media):
            media = client.media(message.sid).list()
            if media:
                for pic in media:
                    item["images"].append(str(pic.uri))
        output["texts"].append(item)
        message.delete()
    ret.append(output)
    return ret



"""
Validate the boto_sns module
"""
import re

import pytest

from tests.support.case import ModuleCase

try:
    import boto

    NO_BOTO_MODULE = False
except ImportError:
    NO_BOTO_MODULE = True


@pytest.mark.skipif(
    NO_BOTO_MODULE,
    reason="Please install the boto library before running boto integration tests.",
)
class BotoSNSTest(ModuleCase):
    def setUp(self):
        try:
            boto.connect_iam()
        except boto.exception.NoAuthHandlerFound:
            self.skipTest(
                "Please setup boto AWS credentials before running boto integration"
                " tests."
            )
        # The name of the topic you want to create.
        # Constraints: Topic names must be made up of only uppercase and
        # lowercase ASCII letters, numbers, underscores, and hyphens,
        # and must be between 1 and 256 characters long.
        # http://docs.aws.amazon.com/sns/latest/api/API_CreateTopic.html
        self.topic_name = re.sub(r"[^a-zA-Z_-]", "_", self.id())[0:256]
        self.topic_names = [self.topic_name]
        self.run_function("boto_sns.delete", name=self.topic_name)

    def tearDown(self):
        for topic in self.topic_names:
            self.run_function("boto_sns.delete", name=topic)

    def test_exists_non_existing(self):
        ret = self.run_function("boto_sns.exists", ["nonexistent"])
        self.assertSaltModuleFalseReturn(ret)

    def test_exists_existing(self):
        self.run_function("boto_sns.create", [self.topic_name])
        ret = self.run_function("boto_sns.exists", [self.topic_name])
        self.assertSaltModuleTrueReturn(ret)

    def test_create(self):
        ret = self.run_function("boto_sns.create", [self.topic_name])
        self.assertSaltModuleTrueReturn(ret)

        ret = self.run_function("boto_sns.get_all_topics")
        self.assertIn(self.topic_name, list(ret.keys()))
        self.assertIn(self._get_arn(self.topic_name), list(ret.values()))

    def test_delete_non_existing(self):
        ret = self.run_function("boto_sns.delete", [self.topic_name])
        self.assertSaltModuleTrueReturn(ret)

    def test_delete_existing(self):
        self.run_function("boto_sns.create", [self.topic_name])
        ret = self.run_function("boto_sns.delete", [self.topic_name])
        self.assertSaltModuleTrueReturn(ret)

        ret = self.run_function("boto_sns.get_all_topics")
        self.assertNotIn(self.topic_name, list(ret.keys()))
        self.assertNotIn(self._get_arn(self.topic_name), list(ret.values()))

    def test_get_all_topics(self):
        self.topic_names.append(self.topic_name + "-2")
        for topic in self.topic_names:
            self.run_function("boto_sns.create", [topic])

        ret = self.run_function("boto_sns.get_all_topics")

        for topic in self.topic_names:
            self.assertIn(topic, list(ret.keys()))
            self.assertIn(self._get_arn(topic), list(ret.values()))

    def test_subscribe_and_get_all_subscriptions_by_topic(self):
        topic_name = self.topic_name
        ret = self.run_function("boto_sns.create", [topic_name])

        ret = self.run_function(
            "boto_sns.subscribe",
            [topic_name, "https", "https://www.example.com/sns/endpoint"],
        )
        self.assertSaltModuleTrueReturn(ret)

        ret = self.run_function("boto_sns.get_all_subscriptions_by_topic", [topic_name])
        self.assertDictContainsSubset(
            {"Protocol": "https", "Endpoint": "https://www.example.com/sns/endpoint"},
            ret[0],
        )

    def _get_arn(self, name):
        return "arn:aws:sns:us-east-1:{}:{}".format(self.account_id, name)

    @property
    def account_id(self):
        if not hasattr(self, "_account_id"):
            account_id = self.run_function("boto_iam.get_account_id")
            setattr(self, "_account_id", account_id)
        return self._account_id

    def assertSaltModuleTrueReturn(self, ret):
        self.assertIsInstance(ret, bool)
        self.assertTrue(ret)

    def assertSaltModuleFalseReturn(self, ret):
        self.assertIsInstance(ret, bool)
        self.assertFalse(ret)



"""
Support for Linux File Access Control Lists

The Linux ACL module requires the `getfacl` and `setfacl` binaries.

"""

import salt.utils.path
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "acl"


def __virtual__():
    """
    Only load the module if getfacl is installed
    """
    if salt.utils.path.which("getfacl"):
        return __virtualname__
    return (
        False,
        "The linux_acl execution module cannot be loaded: the getfacl binary is not in"
        " the path.",
    )


def version():
    """
    Return facl version from getfacl --version

    CLI Example:

    .. code-block:: bash

        salt '*' acl.version
    """
    cmd = "getfacl --version"
    out = __salt__["cmd.run"](cmd).splitlines()
    ret = out[0].split()
    return ret[1].strip()


def _raise_on_no_files(*args):
    if not args:
        raise CommandExecutionError(
            "You need to specify at least one file or directory to work with!"
        )


def getfacl(*args, **kwargs):
    """
    Return (extremely verbose) map of FACLs on specified file(s)

    CLI Examples:

    .. code-block:: bash

        salt '*' acl.getfacl /tmp/house/kitchen
        salt '*' acl.getfacl /tmp/house/kitchen /tmp/house/livingroom
        salt '*' acl.getfacl /tmp/house/kitchen /tmp/house/livingroom recursive=True
    """
    recursive = kwargs.pop("recursive", False)

    _raise_on_no_files(*args)

    ret = {}
    cmd = "getfacl --absolute-names"
    if recursive:
        cmd += " -R"
    for dentry in args:
        cmd += ' "{}"'.format(dentry)
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    dentry = ""
    for line in out:
        if not line:
            continue
        elif line.startswith("getfacl"):
            continue
        elif line.startswith("#"):
            comps = line.replace("# ", "").split(": ")
            if comps[0] == "file":
                dentry = comps[1]
                ret[dentry] = {"comment": {}, "user": [], "group": []}
            ret[dentry]["comment"][comps[0]] = comps[1]
            if comps[0] == "flags":
                flags = list(comps[1])
                if flags[0] == "s":
                    ret[dentry]["suid"] = True
                if flags[1] == "s":
                    ret[dentry]["sgid"] = True
                if flags[2] == "t":
                    ret[dentry]["sticky"] = True
        else:
            vals = _parse_acl(
                acl=line,
                user=ret[dentry]["comment"]["owner"],
                group=ret[dentry]["comment"]["group"],
            )
            acl_type = vals["type"]
            del vals["type"]
            for entity in ("user", "group"):
                if entity in vals:
                    usergroup = vals[entity]
                    del vals[entity]
                    if acl_type == "acl":
                        ret[dentry][entity].append({usergroup: vals})
                    elif acl_type == "default":
                        if "defaults" not in ret[dentry]:
                            ret[dentry]["defaults"] = {}
                        if entity not in ret[dentry]["defaults"]:
                            ret[dentry]["defaults"][entity] = []
                        ret[dentry]["defaults"][entity].append({usergroup: vals})
            for entity in ("other", "mask"):
                if entity in vals:
                    del vals[entity]
                    if acl_type == "acl":
                        ret[dentry][entity] = [{"": vals}]
                    elif acl_type == "default":
                        if "defaults" not in ret[dentry]:
                            ret[dentry]["defaults"] = {}
                        ret[dentry]["defaults"][entity] = [{"": vals}]

    return ret


def _parse_acl(acl, user, group):
    """
    Parse a single ACL rule
    """
    comps = acl.split(":")
    vals = {}

    # What type of rule is this?
    vals["type"] = "acl"
    if comps[0] == "default":
        vals["type"] = "default"
        comps.pop(0)

    # If a user is not specified, use the owner of the file
    if comps[0] == "user" and not comps[1]:
        comps[1] = user
    elif comps[0] == "group" and not comps[1]:
        comps[1] = group
    vals[comps[0]] = comps[1]

    # Set the permissions fields
    octal = 0
    vals["permissions"] = {}
    if "r" in comps[-1]:
        octal += 4
        vals["permissions"]["read"] = True
    else:
        vals["permissions"]["read"] = False
    if "w" in comps[-1]:
        octal += 2
        vals["permissions"]["write"] = True
    else:
        vals["permissions"]["write"] = False
    if "x" in comps[-1]:
        octal += 1
        vals["permissions"]["execute"] = True
    else:
        vals["permissions"]["execute"] = False
    vals["octal"] = octal

    return vals


def wipefacls(*args, **kwargs):
    """
    Remove all FACLs from the specified file(s)

    CLI Examples:

    .. code-block:: bash

        salt '*' acl.wipefacls /tmp/house/kitchen
        salt '*' acl.wipefacls /tmp/house/kitchen /tmp/house/livingroom
        salt '*' acl.wipefacls /tmp/house/kitchen /tmp/house/livingroom recursive=True
    """
    recursive = kwargs.pop("recursive", False)

    _raise_on_no_files(*args)
    cmd = "setfacl -b"
    if recursive:
        cmd += " -R"
    for dentry in args:
        cmd += ' "{}"'.format(dentry)
    __salt__["cmd.run"](cmd, python_shell=False)
    return True


def _acl_prefix(acl_type):
    prefix = ""
    if acl_type.startswith("d"):
        prefix = "d:"
        acl_type = acl_type.replace("default:", "")
        acl_type = acl_type.replace("d:", "")
    if acl_type == "user" or acl_type == "u":
        prefix += "u"
    elif acl_type == "group" or acl_type == "g":
        prefix += "g"
    elif acl_type == "mask" or acl_type == "m":
        prefix += "m"
    return prefix


def modfacl(acl_type, acl_name="", perms="", *args, **kwargs):
    """
    Add or modify a FACL for the specified file(s)

    CLI Examples:

    .. code-block:: bash

        salt '*' acl.modfacl user myuser rwx /tmp/house/kitchen
        salt '*' acl.modfacl default:group mygroup rx /tmp/house/kitchen
        salt '*' acl.modfacl d:u myuser 7 /tmp/house/kitchen
        salt '*' acl.modfacl g mygroup 0 /tmp/house/kitchen /tmp/house/livingroom
        salt '*' acl.modfacl user myuser rwx /tmp/house/kitchen recursive=True
        salt '*' acl.modfacl user myuser rwx /tmp/house/kitchen raise_err=True
    """
    recursive = kwargs.pop("recursive", False)
    raise_err = kwargs.pop("raise_err", False)

    _raise_on_no_files(*args)

    cmd = "setfacl"
    if recursive:
        # -R must come first as -m needs the acl_* arguments that come later
        cmd += " -R"

    cmd += " -m"

    cmd = "{} {}:{}:{}".format(cmd, _acl_prefix(acl_type), acl_name, perms)

    for dentry in args:
        cmd += ' "{}"'.format(dentry)
    __salt__["cmd.run"](cmd, python_shell=False, raise_err=raise_err)
    return True


def delfacl(acl_type, acl_name="", *args, **kwargs):
    """
    Remove specific FACL from the specified file(s)

    CLI Examples:

    .. code-block:: bash

        salt '*' acl.delfacl user myuser /tmp/house/kitchen
        salt '*' acl.delfacl default:group mygroup /tmp/house/kitchen
        salt '*' acl.delfacl d:u myuser /tmp/house/kitchen
        salt '*' acl.delfacl g myuser /tmp/house/kitchen /tmp/house/livingroom
        salt '*' acl.delfacl user myuser /tmp/house/kitchen recursive=True
    """
    recursive = kwargs.pop("recursive", False)

    _raise_on_no_files(*args)

    cmd = "setfacl"
    if recursive:
        cmd += " -R"

    cmd += " -x"

    cmd = "{} {}:{}".format(cmd, _acl_prefix(acl_type), acl_name)

    for dentry in args:
        cmd += ' "{}"'.format(dentry)
    __salt__["cmd.run"](cmd, python_shell=False)
    return True



"""
Tests for win_file execution module
"""
import pytest

import salt.modules.win_file as win_file
import salt.utils.files
import salt.utils.path
import salt.utils.win_dacl
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {
        win_file: {
            "__utils__": {
                "dacl.get_name": salt.utils.win_dacl.get_name,
                "dacl.get_owner": salt.utils.win_dacl.get_owner,
                "dacl.get_primary_group": salt.utils.win_dacl.get_primary_group,
                "dacl.get_sid_string": salt.utils.win_dacl.get_sid_string,
                "files.normalize_mode": salt.utils.files.normalize_mode,
                "path.islink": salt.utils.path.islink,
            }
        }
    }


@pytest.fixture(scope="function")
def test_file():
    with pytest.helpers.temp_file("win_file_test.file") as test_file:
        yield test_file


def test_stat(test_file):
    ret = win_file.stats(str(test_file), None, True)
    assert ret["mode"] == "0666"
    assert ret["type"] == "file"


def test_stats_issue_43328(test_file):
    """
    Make sure that a CommandExecutionError is raised if the file does NOT
    exist
    """
    fake_file = test_file.parent / "fake.file"
    with pytest.raises(CommandExecutionError):
        win_file.stats(fake_file)



"""
    :codeauthor: :email:`Agnes Tevesz (agnes.tevesz@morganstanley.com)`

    salt.config.schemas.esxvm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ESX Virtual Machine configuration schemas
"""


from salt.utils.schema import (
    AnyOfItem,
    ArrayItem,
    BooleanItem,
    ComplexSchemaItem,
    DefinitionsSchema,
    IntegerItem,
    IPv4Item,
    NullItem,
    NumberItem,
    StringItem,
)


class ESXVirtualMachineSerialBackingItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Serial Port Backing
    """

    title = "ESX Virtual Machine Serial Port Backing"
    description = "ESX virtual machine serial port backing"
    required = True

    uri = StringItem()
    direction = StringItem(enum=("client", "server"))
    filename = StringItem()


class ESXVirtualMachineDeviceConnectionItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Serial Port Connection
    """

    title = "ESX Virtual Machine Serial Port Connection"
    description = "ESX virtual machine serial port connection"
    required = True

    allow_guest_control = BooleanItem(default=True)
    start_connected = BooleanItem(default=True)


class ESXVirtualMachinePlacementSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Placement
    """

    title = "ESX Virtual Machine Placement Information"
    description = "ESX virtual machine placement property"
    required = True

    cluster = StringItem(
        title="Virtual Machine Cluster",
        description="Cluster of the virtual machine if it is placed to a cluster",
    )
    host = StringItem(
        title="Virtual Machine Host",
        description="Host of the virtual machine if it is placed to a standalone host",
    )
    resourcepool = StringItem(
        title="Virtual Machine Resource Pool",
        description=(
            "Resource pool of the virtual machine if it is placed to a resource pool"
        ),
    )
    folder = StringItem(
        title="Virtual Machine Folder",
        description=(
            "Folder of the virtual machine where it should be deployed, default is the"
            " datacenter vmFolder"
        ),
    )


class ESXVirtualMachineCdDriveClientSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine CD Drive Client
    """

    title = "ESX Virtual Machine Serial CD Client"
    description = "ESX virtual machine CD/DVD drive client properties"

    mode = StringItem(required=True, enum=("passthrough", "atapi"))


class ESXVirtualMachineCdDriveIsoSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine CD Drive ISO
    """

    title = "ESX Virtual Machine Serial CD ISO"
    description = "ESX virtual machine CD/DVD drive ISO properties"

    path = StringItem(required=True)


class ESXVirtualMachineCdDriveSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine CD Drives
    """

    title = "ESX Virtual Machine Serial CD"
    description = "ESX virtual machine CD/DVD drive properties"

    adapter = StringItem(
        title="Virtual Machine CD/DVD Adapter",
        description="Unique adapter name for virtual machine cd/dvd drive",
        required=True,
    )
    controller = StringItem(required=True)
    device_type = StringItem(
        title="Virtual Machine Device Type",
        description="CD/DVD drive of the virtual machine if it is placed to a cluster",
        required=True,
        default="client_device",
        enum=("datastore_iso_file", "client_device"),
    )
    client_device = ESXVirtualMachineCdDriveClientSchemaItem()
    datastore_iso_file = ESXVirtualMachineCdDriveIsoSchemaItem()
    connectable = ESXVirtualMachineDeviceConnectionItem()


class ESXVirtualMachineSerialSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Serial Port
    """

    title = "ESX Virtual Machine Serial Port Configuration"
    description = "ESX virtual machine serial port properties"

    type = StringItem(
        title="Virtual Machine Serial Port Type",
        required=True,
        enum=("network", "pipe", "file", "device"),
    )
    adapter = StringItem(
        title="Virtual Machine Serial Port Name",
        description=(
            "Unique adapter name for virtual machine serial port "
            "for creation an arbitrary value should be specified"
        ),
        required=True,
    )
    backing = ESXVirtualMachineSerialBackingItem()
    connectable = ESXVirtualMachineDeviceConnectionItem()
    yield_port = BooleanItem(
        title="Serial Port Yield", description="Serial port yield", default=False
    )


class ESXVirtualMachineScsiSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine SCSI Controller
    """

    title = "ESX Virtual Machine SCSI Controller Configuration"
    description = "ESX virtual machine scsi controller properties"
    required = True

    adapter = StringItem(
        title="Virtual Machine SCSI Controller Name",
        description=(
            "Unique SCSI controller name "
            "for creation an arbitrary value should be specified"
        ),
        required=True,
    )
    type = StringItem(
        title="Virtual Machine SCSI type",
        description="Type of the SCSI controller",
        required=True,
        enum=("lsilogic", "lsilogic_sas", "paravirtual", "buslogic"),
    )
    bus_sharing = StringItem(
        title="Virtual Machine SCSI bus sharing",
        description="Sharing type of the SCSI bus",
        required=True,
        enum=("virtual_sharing", "physical_sharing", "no_sharing"),
    )
    bus_number = NumberItem(
        title="Virtual Machine SCSI bus number",
        description="Unique bus number of the SCSI device",
        required=True,
    )


class ESXVirtualMachineSataSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine SATA Controller
    """

    title = "ESX Virtual Machine SATA Controller Configuration"
    description = "ESX virtual machine SATA controller properties"
    required = False
    adapter = StringItem(
        title="Virtual Machine SATA Controller Name",
        description=(
            "Unique SATA controller name "
            "for creation an arbitrary value should be specified"
        ),
        required=True,
    )
    bus_number = NumberItem(
        title="Virtual Machine SATA bus number",
        description="Unique bus number of the SATA device",
        required=True,
    )


class ESXVirtualMachineDiskSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Disk
    """

    title = "ESX Virtual Machine Disk Configuration"
    description = "ESX virtual machine disk properties"
    required = True

    size = NumberItem(
        title="Disk size", description="Size of the disk in GB", required=True
    )
    unit = StringItem(
        title="Disk size unit",
        description=(
            "Unit of the disk size, to VMware a GB is the same as GiB = 1024MiB"
        ),
        required=False,
        default="GB",
        enum=("KB", "MB", "GB"),
    )
    adapter = StringItem(
        title="Virtual Machine Adapter Name",
        description=(
            "Unique adapter name for virtual machine "
            "for creation an arbitrary value should be specified"
        ),
        required=True,
    )
    filename = StringItem(
        title="Virtual Machine Disk File",
        description="File name of the virtual machine vmdk",
    )
    datastore = StringItem(
        title="Virtual Machine Disk Datastore",
        description="Disk datastore where the virtual machine files will be placed",
        required=True,
    )
    address = StringItem(
        title="Virtual Machine SCSI Address",
        description="Address of the SCSI adapter for the virtual machine",
        pattern=r"\d:\d",
    )
    thin_provision = BooleanItem(
        title="Virtual Machine Disk Provision Type",
        description="Provision type of the disk",
        default=True,
        required=False,
    )
    eagerly_scrub = AnyOfItem(required=False, items=[BooleanItem(), NullItem()])
    controller = StringItem(
        title="Virtual Machine SCSI Adapter",
        description="Name of the SCSI adapter where the disk will be connected",
        required=True,
    )


class ESXVirtualMachineNicMapSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Nic Map
    """

    title = "ESX Virtual Machine Nic Configuration"
    description = "ESX Virtual Machine nic properties"
    required = False

    domain = StringItem()
    gateway = IPv4Item()
    ip_addr = IPv4Item()
    subnet_mask = IPv4Item()


class ESXVirtualMachineInterfaceSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Network Interface
    """

    title = "ESX Virtual Machine Network Interface Configuration"
    description = "ESX Virtual Machine network adapter properties"
    required = True

    name = StringItem(
        title="Virtual Machine Port Group",
        description="Specifies the port group name for the virtual machine connection",
        required=True,
    )
    adapter = StringItem(
        title="Virtual Machine Network Adapter",
        description=(
            "Unique name of the network adapter, "
            "for creation an arbitrary value should be specified"
        ),
        required=True,
    )
    adapter_type = StringItem(
        title="Virtual Machine Adapter Type",
        description="Network adapter type of the virtual machine",
        required=True,
        enum=("vmxnet", "vmxnet2", "vmxnet3", "e1000", "e1000e"),
        default="vmxnet3",
    )
    switch_type = StringItem(
        title="Virtual Machine Switch Type",
        description=(
            "Specifies the type of the virtual switch for the virtual machine"
            " connection"
        ),
        required=True,
        default="standard",
        enum=("standard", "distributed"),
    )
    mac = StringItem(
        title="Virtual Machine MAC Address",
        description="Mac address of the virtual machine",
        required=False,
        pattern="^([0-9a-f]{1,2}[:]){5}([0-9a-f]{1,2})$",
    )
    mapping = ESXVirtualMachineNicMapSchemaItem()
    connectable = ESXVirtualMachineDeviceConnectionItem()


class ESXVirtualMachineMemorySchemaItem(ComplexSchemaItem):
    """
    Configurtation Schema Item for ESX Virtual Machine Memory
    """

    title = "ESX Virtual Machine Memory Configuration"
    description = "ESX Virtual Machine memory property"
    required = True

    size = IntegerItem(
        title="Memory size", description="Size of the memory", required=True
    )

    unit = StringItem(
        title="Memory unit",
        description="Unit of the memory, to VMware a GB is the same as GiB = 1024MiB",
        required=False,
        default="MB",
        enum=("MB", "GB"),
    )
    hotadd = BooleanItem(required=False, default=False)
    reservation_max = BooleanItem(required=False, default=False)


class ESXVirtualMachineCpuSchemaItem(ComplexSchemaItem):
    """
    Configurtation Schema Item for ESX Virtual Machine CPU
    """

    title = "ESX Virtual Machine Memory Configuration"
    description = "ESX Virtual Machine memory property"
    required = True

    count = IntegerItem(
        title="CPU core count", description="CPU core count", required=True
    )
    cores_per_socket = IntegerItem(
        title="CPU cores per socket",
        description="CPU cores per socket count",
        required=False,
    )
    nested = BooleanItem(
        title="Virtual Machine Nested Property",
        description="Nested virtualization support",
        default=False,
    )
    hotadd = BooleanItem(
        title="Virtual Machine CPU hot add", description="CPU hot add", default=False
    )
    hotremove = BooleanItem(
        title="Virtual Machine CPU hot remove",
        description="CPU hot remove",
        default=False,
    )


class ESXVirtualMachineConfigSchema(DefinitionsSchema):
    """
    Configuration Schema for ESX Virtual Machines
    """

    title = "ESX Virtual Machine Configuration Schema"
    description = "ESX Virtual Machine configuration schema"

    vm_name = StringItem(
        title="Virtual Machine name",
        description="Name of the virtual machine",
        required=True,
    )
    cpu = ESXVirtualMachineCpuSchemaItem()
    memory = ESXVirtualMachineMemorySchemaItem()
    image = StringItem(
        title="Virtual Machine guest OS", description="Guest OS type", required=True
    )
    version = StringItem(
        title="Virtual Machine hardware version",
        description="Container hardware version property",
        required=True,
    )
    interfaces = ArrayItem(
        items=ESXVirtualMachineInterfaceSchemaItem(),
        min_items=1,
        required=False,
        unique_items=True,
    )
    disks = ArrayItem(
        items=ESXVirtualMachineDiskSchemaItem(),
        min_items=1,
        required=False,
        unique_items=True,
    )
    scsi_devices = ArrayItem(
        items=ESXVirtualMachineScsiSchemaItem(),
        min_items=1,
        required=False,
        unique_items=True,
    )
    serial_ports = ArrayItem(
        items=ESXVirtualMachineSerialSchemaItem(),
        min_items=0,
        required=False,
        unique_items=True,
    )
    cd_dvd_drives = ArrayItem(
        items=ESXVirtualMachineCdDriveSchemaItem(),
        min_items=0,
        required=False,
        unique_items=True,
    )
    sata_controllers = ArrayItem(
        items=ESXVirtualMachineSataSchemaItem(),
        min_items=0,
        required=False,
        unique_items=True,
    )
    datacenter = StringItem(
        title="Virtual Machine Datacenter",
        description="Datacenter of the virtual machine",
        required=True,
    )
    datastore = StringItem(
        title="Virtual Machine Datastore",
        description="Datastore of the virtual machine",
        required=True,
    )
    placement = ESXVirtualMachinePlacementSchemaItem()
    template = BooleanItem(
        title="Virtual Machine Template",
        description="Template to create the virtual machine from",
        default=False,
    )
    tools = BooleanItem(
        title="Virtual Machine VMware Tools",
        description="Install VMware tools on the guest machine",
        default=False,
    )
    power_on = BooleanItem(
        title="Virtual Machine Power",
        description="Power on virtual machine afret creation",
        default=False,
    )
    deploy = BooleanItem(
        title="Virtual Machine Deploy Salt",
        description="Deploy salt after successful installation",
        default=False,
    )


class ESXVirtualMachineRemoveSchema(DefinitionsSchema):
    """
    Remove Schema for ESX Virtual Machines to delete or unregister virtual machines
    """

    name = StringItem(
        title="Virtual Machine name",
        description="Name of the virtual machine",
        required=True,
    )
    datacenter = StringItem(
        title="Virtual Machine Datacenter",
        description="Datacenter of the virtual machine",
        required=True,
    )
    placement = AnyOfItem(
        required=False, items=[ESXVirtualMachinePlacementSchemaItem(), NullItem()]
    )
    power_off = BooleanItem(
        title="Power off vm",
        description="Power off vm before delete operation",
        required=False,
    )


class ESXVirtualMachineDeleteSchema(ESXVirtualMachineRemoveSchema):
    """
    Deletion Schema for ESX Virtual Machines
    """


class ESXVirtualMachineUnregisterSchema(ESXVirtualMachineRemoveSchema):
    """
    Unregister Schema for ESX Virtual Machines
    """



"""
Tests for the archive state
"""
import os
import pathlib
import shutil
import tempfile
import textwrap

import attr
import pytest

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils

try:
    import zipfile  # pylint: disable=unused-import

    HAS_ZIPFILE = True
except ImportError:
    HAS_ZIPFILE = False

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@attr.s(frozen=True, slots=True)
class Archive:
    fmt = attr.ib()
    unicode_filename = attr.ib(default=False)

    path = attr.ib(init=False)
    src = attr.ib(init=False)
    src_file = attr.ib(init=False)
    archive = attr.ib(init=False)
    dst = attr.ib(init=False)
    filename = attr.ib(init=False)

    @path.default
    def _path(self):
        return pathlib.Path(tempfile.mkdtemp())

    @src.default
    def _src(self):
        return self.path / "{}_src_dir".format(self.fmt)

    @src_file.default
    def _src_file(self):
        return self.src / "file"

    @archive.default
    def _archive(self):
        return self.path / "archive.{}".format(self.fmt)

    @dst.default
    def _dst(self):
        return self.path / "{}_dst_dir".format(self.fmt)

    @filename.default
    def _filename(self):
        if self.unicode_filename:
            return "file®"
        return "file"

    def __attrs_post_init__(self):
        # Create source
        self.src.mkdir()
        self.dst.mkdir()

        if salt.utils.platform.is_windows():
            encoding = "utf-8"
        else:
            encoding = None

        dst_filename = self.src / self.filename
        dst_filename.write_bytes(
            salt.utils.stringutils.to_bytes(
                textwrap.dedent(
                    """\
            Compression theorem of computational complexity theory:

            Given a Gödel numbering $φ$ of the computable functions and a
            Blum complexity measure $Φ$ where a complexity class for a
            boundary function $f$ is defined as

                $\\mathrm C(f) := \\{φ_i ∈ \\mathbb R^{(1)} | (∀^∞ x) Φ_i(x) ≤ f(x)\\}$.

            Then there exists a total computable function $f$ so that for
            all $i$

                $\\mathrm{Dom}(φ_i) = \\mathrm{Dom}(φ_{f(i)})$

            and

                $\\mathrm C(φ_i) ⊊ \\mathrm{C}(φ_{f(i)})$.
        """
                ),
                encoding=encoding,
            )
        )

    def assert_artifacts_in_ret(self, ret, file_only=False, unix_sep=False):
        """
        Assert that the artifact source files are printed in the source command
        output
        """

        def normdir(path):
            normdir = os.path.normcase(os.path.abspath(str(path)))
            if salt.utils.platform.is_windows():
                # Remove the drive portion of path
                if len(normdir) >= 2 and normdir[1] == ":":
                    normdir = normdir.split(":", 1)[1]
            normdir = normdir.lstrip(os.path.sep)
            # Unzipped paths might have unix line endings
            if unix_sep:
                normdir = normdir.replace(os.path.sep, "/")
            return normdir

        # Try to find source directory and file in output lines
        dir_in_ret = None
        file_in_ret = None
        for line in ret:
            if normdir(self.src) in os.path.normcase(line) and not normdir(
                self.src_file
            ) in os.path.normcase(line):
                dir_in_ret = True
            if normdir(self.src_file) in os.path.normcase(line):
                file_in_ret = True

        # Assert number of lines, reporting of source directory and file
        assert len(ret) >= 1 if file_only else 2
        if not file_only:
            assert dir_in_ret is True
        assert file_in_ret is True

    def __enter__(self):
        return self

    def __exit__(self, *_):
        shutil.rmtree(str(self.path), ignore_errors=True)


@pytest.fixture(scope="module")
def archive(modules):
    return modules.archive


def unicode_filename_ids(value):
    return "unicode_filename={}".format(value)


@pytest.fixture(params=[True, False], ids=unicode_filename_ids)
def unicode_filename(request):
    return request.param


@pytest.mark.skip_on_windows
@pytest.mark.skip_if_binaries_missing("tar")
def test_tar_pack(archive, unicode_filename):
    """
    Validate using the tar function to create archives
    """
    with Archive("tar", unicode_filename=unicode_filename) as arch:
        ret = archive.tar("-cvf", str(arch.archive), sources=str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skip_on_windows
@pytest.mark.skip_if_binaries_missing("tar")
def test_tar_unpack(archive, unicode_filename):
    """
    Validate using the tar function to extract archives
    """
    with Archive("tar", unicode_filename=unicode_filename) as arch:
        ret = archive.tar("-cvf", str(arch.archive), sources=str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)

        ret = archive.tar("-xvf", str(arch.archive), dest=str(arch.dst))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skip_on_windows
@pytest.mark.skip_if_binaries_missing("tar")
def test_tar_list(archive, unicode_filename):
    """
    Validate using the tar function to list archives
    """
    with Archive("tar", unicode_filename=unicode_filename) as arch:
        ret = archive.tar("-cvf", str(arch.archive), sources=str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)

        ret = archive.list(str(arch.archive))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skip_if_binaries_missing("gzip")
def test_gzip(archive, unicode_filename):
    """
    Validate using the gzip function
    """
    with Archive("gz", unicode_filename=unicode_filename) as arch:
        ret = archive.gzip(str(arch.src_file), options="-v")
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret, file_only=True)


@pytest.mark.skip_if_binaries_missing("gzip", "gunzip")
def test_gunzip(archive, unicode_filename):
    """
    Validate using the gunzip function
    """
    with Archive("gz", unicode_filename=unicode_filename) as arch:
        ret = archive.gzip(str(arch.src_file), options="-v")
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret, file_only=True)

        ret = archive.gunzip(str(arch.src_file) + ".gz", options="-v")
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret, file_only=True)


@pytest.mark.skip_if_binaries_missing("zip")
def test_cmd_zip(archive, unicode_filename, grains):
    """
    Validate using the cmd_zip function
    """
    with Archive("zip", unicode_filename=unicode_filename) as arch:
        ret = archive.cmd_zip(str(arch.archive), str(arch.src))
        assert isinstance(ret, list)
        if (
            "*** buffer overflow detected ***: terminated" in ret
            and grains["os"] == "Fedora"
            and int(grains["osmajorrelease"]) >= 37
        ):
            pytest.skip(
                f"Calling the system zip on {grains['os']} {grains['osmajorrelease']} "
                "is currently failing with a buffer overflow error"
            )
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skip_if_binaries_missing("zip", "unzip")
def test_cmd_unzip(archive, unicode_filename, grains):
    """
    Validate using the cmd_unzip function
    """
    with Archive("zip", unicode_filename=unicode_filename) as arch:
        ret = archive.cmd_zip(str(arch.archive), str(arch.src))
        assert isinstance(ret, list)
        if (
            "*** buffer overflow detected ***: terminated" in ret
            and grains["os"] == "Fedora"
            and int(grains["osmajorrelease"]) >= 37
        ):
            pytest.skip(
                f"Calling the system unzip on {grains['os']} {grains['osmajorrelease']} "
                "is currently failing with a buffer overflow error"
            )
        arch.assert_artifacts_in_ret(ret)

        ret = archive.cmd_unzip(str(arch.archive), str(arch.dst))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skipif(not HAS_ZIPFILE, reason="Cannot find zipfile python module")
def test_zip(archive, unicode_filename):
    """
    Validate using the zip function
    """
    with Archive("zip", unicode_filename=unicode_filename) as arch:
        ret = archive.zip(str(arch.archive), str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skipif(not HAS_ZIPFILE, reason="Cannot find zipfile python module")
def test_unzip(archive, unicode_filename):
    """
    Validate using the unzip function
    """
    with Archive("zip", unicode_filename=unicode_filename) as arch:
        ret = archive.zip(str(arch.archive), str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)

        ret = archive.unzip(str(arch.archive), str(arch.dst))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret, unix_sep=False)


@pytest.mark.skip_if_binaries_missing("rar")
def test_rar(archive, unicode_filename):
    """
    Validate using the rar function
    """
    with Archive("rar", unicode_filename=unicode_filename) as arch:
        ret = archive.rar(str(arch.archive), str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skip_if_binaries_missing("rar", "unrar")
def test_unrar(archive, unicode_filename):
    """
    Validate using the unrar function
    """
    with Archive("rar", unicode_filename=unicode_filename) as arch:
        ret = archive.rar(str(arch.archive), str(arch.src))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)

        ret = archive.unrar(str(arch.archive), str(arch.dst))
        assert isinstance(ret, list)
        arch.assert_artifacts_in_ret(ret)


@pytest.mark.skip_on_windows
@pytest.mark.skip_if_binaries_missing("tar")
def test_tar_list_no_explicit_top_level_directory_member(archive, tmp_path):
    rel_path = pathlib.Path("archive_top/second_level/file.txt")
    srcfile = tmp_path / str(rel_path)
    srcfile.parent.mkdir(parents=True, exist_ok=True)
    srcfile.write_text("hey there")
    archive.tar(
        "-cvzf",
        str(tmp_path / "tld_test.tar.gz"),
        sources=str(rel_path.parent),
        cwd=str(tmp_path),
    )

    expected = {
        "dirs": ["archive_top/second_level/"],
        "files": ["archive_top/second_level/file.txt"],
        "links": [],
        "top_level_dirs": ["archive_top/"],
        "top_level_files": [],
        "top_level_links": [],
    }
    ret = archive.list(
        str(tmp_path / "tld_test.tar.gz"), archive_format="tar", verbose=True
    )
    assert ret == expected


@pytest.mark.skip_on_windows
@pytest.mark.skip_if_binaries_missing("tar")
def test_tar_list_with_similar_top_level_dirs(archive, tmp_path):
    rel_paths = (
        "archive_top/second_level/file.txt",
        "archives/second_level/file.txt",
    )

    for rel_path in rel_paths:
        srcfile = tmp_path / str(pathlib.Path(rel_path))
        srcfile.parent.mkdir(parents=True, exist_ok=True)
        srcfile.write_text("hey there")

    archive.tar(
        "-cvzf",
        str(tmp_path / "tld_test.tar.gz"),
        sources=["archive_top", "archives"],
        cwd=str(tmp_path),
    )

    expected = {
        "dirs": [
            "archive_top/",
            "archive_top/second_level/",
            "archives/",
            "archives/second_level/",
        ],
        "files": [
            "archive_top/second_level/file.txt",
            "archives/second_level/file.txt",
        ],
        "links": [],
        "top_level_dirs": ["archive_top/", "archives/"],
        "top_level_files": [],
        "top_level_links": [],
    }
    ret = archive.list(
        str(tmp_path / "tld_test.tar.gz"), archive_format="tar", verbose=True
    )
    assert ret == expected



"""
Cache data in filesystem.

.. versionadded:: 2016.11.0

The ``localfs`` Minion cache module is the default cache module and does not
require any configuration.

Expiration values can be set in the relevant config file (``/etc/salt/master`` for
the master, ``/etc/salt/cloud`` for Salt Cloud, etc).
"""

import errno
import logging
import os
import os.path
import shutil
import tempfile

import salt.payload
import salt.utils.atomicfile
import salt.utils.files
from salt.exceptions import SaltCacheError

log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}


def __cachedir(kwargs=None):
    if kwargs and "cachedir" in kwargs:
        return kwargs["cachedir"]
    return __opts__.get("cachedir", salt.syspaths.CACHE_DIR)


def init_kwargs(kwargs):
    return {"cachedir": __cachedir(kwargs)}


def get_storage_id(kwargs):
    return ("localfs", __cachedir(kwargs))


def store(bank, key, data, cachedir):
    """
    Store information in a file.
    """
    base = os.path.join(cachedir, os.path.normpath(bank))
    try:
        os.makedirs(base)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise SaltCacheError(
                "The cache directory, {}, could not be created: {}".format(base, exc)
            )

    outfile = os.path.join(base, "{}.p".format(key))
    tmpfh, tmpfname = tempfile.mkstemp(dir=base)
    os.close(tmpfh)
    try:
        with salt.utils.files.fopen(tmpfname, "w+b") as fh_:
            salt.payload.dump(data, fh_)
        # On Windows, os.rename will fail if the destination file exists.
        salt.utils.atomicfile.atomic_rename(tmpfname, outfile)
    except OSError as exc:
        raise SaltCacheError(
            "There was an error writing the cache file, {}: {}".format(base, exc)
        )


def fetch(bank, key, cachedir):
    """
    Fetch information from a file.
    """
    inkey = False
    key_file = os.path.join(cachedir, os.path.normpath(bank), "{}.p".format(key))
    if not os.path.isfile(key_file):
        # The bank includes the full filename, and the key is inside the file
        key_file = os.path.join(cachedir, os.path.normpath(bank) + ".p")
        inkey = True

    if not os.path.isfile(key_file):
        log.debug('Cache file "%s" does not exist', key_file)
        return {}
    try:
        with salt.utils.files.fopen(key_file, "rb") as fh_:
            if inkey:
                return salt.payload.load(fh_)[key]
            else:
                return salt.payload.load(fh_)
    except OSError as exc:
        raise SaltCacheError(
            'There was an error reading the cache file "{}": {}'.format(key_file, exc)
        )


def updated(bank, key, cachedir):
    """
    Return the epoch of the mtime for this cache file
    """
    key_file = os.path.join(cachedir, os.path.normpath(bank), "{}.p".format(key))
    if not os.path.isfile(key_file):
        log.warning('Cache file "%s" does not exist', key_file)
        return None
    try:
        return int(os.path.getmtime(key_file))
    except OSError as exc:
        raise SaltCacheError(
            'There was an error reading the mtime for "{}": {}'.format(key_file, exc)
        )


def flush(bank, key=None, cachedir=None):
    """
    Remove the key from the cache bank with all the key content.
    """
    if cachedir is None:
        cachedir = __cachedir()

    try:
        if key is None:
            target = os.path.join(cachedir, os.path.normpath(bank))
            if not os.path.isdir(target):
                return False
            shutil.rmtree(target)
        else:
            target = os.path.join(cachedir, os.path.normpath(bank), "{}.p".format(key))
            if not os.path.isfile(target):
                return False
            os.remove(target)
    except OSError as exc:
        raise SaltCacheError('There was an error removing "{}": {}'.format(target, exc))
    return True


def list_(bank, cachedir):
    """
    Return an iterable object containing all entries stored in the specified bank.
    """
    base = os.path.join(cachedir, os.path.normpath(bank))
    if not os.path.isdir(base):
        return []
    try:
        items = os.listdir(base)
    except OSError as exc:
        raise SaltCacheError(
            'There was an error accessing directory "{}": {}'.format(base, exc)
        )
    ret = []
    for item in items:
        if item.endswith(".p"):
            ret.append(item.rstrip(item[-2:]))
        else:
            ret.append(item)
    return ret


def contains(bank, key, cachedir):
    """
    Checks if the specified bank contains the specified key.
    """
    if key is None:
        base = os.path.join(cachedir, os.path.normpath(bank))
        return os.path.isdir(base)
    else:
        keyfile = os.path.join(cachedir, os.path.normpath(bank), "{}.p".format(key))
        return os.path.isfile(keyfile)



"""
Commands for working with travisci.

:depends: pyOpenSSL >= 16.0.0
"""
import base64
import urllib.parse

import salt.utils.json
from salt.utils.versions import Version

try:
    import OpenSSL
    import OpenSSL.crypto

    HAS_OPENSSL = True
except ImportError:
    HAS_OPENSSL = False


OPENSSL_MIN_VER = "16.0.0"
__virtualname__ = "travisci"


def __virtual__():
    if HAS_OPENSSL is False:
        return (
            False,
            "The travisci module was unable to be loaded: Install pyOpenssl >= {}".format(
                OPENSSL_MIN_VER
            ),
        )
    cur_version = Version(OpenSSL.__version__)
    min_version = Version(OPENSSL_MIN_VER)
    if cur_version < min_version:
        return (
            False,
            "The travisci module was unable to be loaded: Install pyOpenssl >= {}".format(
                OPENSSL_MIN_VER
            ),
        )
    return __virtualname__


def verify_webhook(signature, body):
    """
    Verify the webhook signature from travisci

    signature
        The signature header from the webhook header

    body
        The full payload body from the webhook post

    .. note:: The body needs to be the urlencoded version of the body.

    CLI Example:

    .. code-block:: bash

        salt '*' travisci.verify_webhook 'M6NucCX5722bxisQs7e...' 'payload=%7B%22id%22%3A183791261%2C%22repository...'

    """
    # get public key setup
    public_key = __utils__["http.query"]("https://api.travis-ci.org/config")["config"][
        "notifications"
    ]["webhook"]["public_key"]
    pkey_public_key = OpenSSL.crypto.load_publickey(
        OpenSSL.crypto.FILETYPE_PEM, public_key
    )
    certificate = OpenSSL.crypto.X509()
    certificate.set_pubkey(pkey_public_key)

    # decode signature
    signature = base64.b64decode(signature)

    # parse the urlencoded payload from travis
    payload = salt.utils.json.loads(urllib.parse.parse_qs(body)["payload"][0])

    try:
        OpenSSL.crypto.verify(certificate, signature, payload, "sha1")
    except OpenSSL.crypto.Error:
        return False
    return True



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


FIRST_IF_CONTENTS = """\
# set variable identifying the chroot you work in (used in the prompt below)
if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi

"""
SECOND_IF_CONTENTS = """\
# enable bash completion in interactive shells
if [ -f /etc/bash_completion ] && ! shopt -oq posix; then
    . /etc/bash_completion
fi
"""


@pytest.mark.parametrize("test", (True, False))
def test_append(file, tmp_path, test):
    """
    file.append
    """
    name = tmp_path / "testfile"
    name.write_text("#salty!")
    ret = file.append(name=str(name), text="cheese", test=test)
    if test is True:
        assert ret.result is None
        assert name.read_text() == "#salty!"
    else:
        assert ret.result is True
        assert name.read_text() == "#salty!\ncheese\n"


def test_append_issue_1864_makedirs(file, tmp_path):
    """
    file.append but create directories if needed as an option, and create
    the file if it doesn't exist
    """
    fname = "append_issue_1864_makedirs"
    name = tmp_path / fname

    # Non existing file get's touched
    ret = file.append(name=str(name), text="cheese", makedirs=True)
    assert ret.result is True
    assert name.is_file()

    # Nested directory and file get's touched
    name = tmp_path / "issue_1864" / fname
    ret = file.append(name=str(name), text="cheese", makedirs=True)
    assert ret.result is True
    assert name.is_file()
    assert name.parent.is_dir()

    # Parent directory exists but file does not and makedirs is False
    name = name.with_name(name.name + "2")
    ret = file.append(name=str(name), text="cheese", makedirs=False)
    assert ret.result is True
    assert name.is_file()


def test_issue_2227_file_append(file, tmp_path):
    """
    Text to append includes a percent symbol
    """
    # let's make use of existing state to create a file with contents to
    # test against
    tmp_file_append = tmp_path / "test.append"
    tmp_file_append.write_text(FIRST_IF_CONTENTS + SECOND_IF_CONTENTS)
    ret = file.append(name=str(tmp_file_append), text="HISTTIMEFORMAT='%F %T '")
    assert ret.result is True
    contents_pre = tmp_file_append.read_text()

    # It should not append text again
    ret = file.append(name=str(tmp_file_append), text="HISTTIMEFORMAT='%F %T '")
    assert ret.result is True
    contents_post = tmp_file_append.read_text()

    assert contents_pre == contents_post


def test_issue_2379_file_append(modules, tmp_path):
    # Get a path to the temporary file
    tmp_file = tmp_path / "issue-2379-file-append.txt"
    # Write some data to it
    tmp_file.write_text(
        "hello\nworld\n"  # Some junk
        "#PermitRootLogin yes\n"  # Commented text
        "# PermitRootLogin yes\n"  # Commented text with space
    )
    # create the sls template
    template_lines = [
        "{}:".format(tmp_file),
        "  file.append:",
        "    - text: PermitRootLogin yes",
    ]
    template = "\n".join(template_lines)
    ret = modules.state.template_str(template)
    for state_run in ret:
        assert state_run.result is True
        assert "Appended 1 lines" in state_run.comment


@pytest.mark.slow_test
def test_issue_1896_file_append_source(file, tmp_path, state_tree):
    """
    Verify that we can append a file's contents
    """
    testfile = tmp_path / "test.append"
    testfile.touch()

    firstif_file = pytest.helpers.temp_file(
        "firstif", directory=state_tree / "testappend", contents=FIRST_IF_CONTENTS
    )
    secondif_file = pytest.helpers.temp_file(
        "secondif", directory=state_tree / "testappend", contents=SECOND_IF_CONTENTS
    )
    with firstif_file, secondif_file:
        ret = file.append(name=str(testfile), source="salt://testappend/firstif")
        assert ret.result is True
        ret = file.append(name=str(testfile), source="salt://testappend/secondif")
        assert ret.result is True

        testfile_contents = testfile.read_text()

        assert testfile_contents == FIRST_IF_CONTENTS + SECOND_IF_CONTENTS

        # Run it again
        ret = file.append(name=str(testfile), source="salt://testappend/firstif")
        assert ret.result is True
        ret = file.append(name=str(testfile), source="salt://testappend/secondif")
        assert ret.result is True

        testfile_contents = testfile.read_text()

        assert testfile_contents == FIRST_IF_CONTENTS + SECOND_IF_CONTENTS


def test_file_append_check_cmd(modules, state_tree, tmp_path):
    """
    Test that check_cmd works for file.append
    and those states do not run.
    """
    sls_contents = f"""
append_in_file:
  file.append:
    - name: /tmp/test
    - text: "appended text"
    - check_cmd:
      - "djasjahj"
    """
    with pytest.helpers.temp_file(
        "file-append-check-cmd.sls", sls_contents, state_tree
    ):
        ret = modules.state.sls("file-append-check-cmd")
        for state_run in ret:
            assert state_run.result is False
            assert state_run.comment == "check_cmd determined the state failed"
